#encoding: UTF-8
=begin
/***************************************************************************
 *   ©2011-2014 Michael Uplawski <michael.uplawski@uplawski.eu>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         * 
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
=end

require_relative 'file_checking'
require 'singleton'
require 'yaml'
require_relative 'logging'

# An object of the Configuration class reads and exposes options from a 
# configuration-file in YAML format.
class Configuration
  include File_Checking
  include Singleton
  include Logging
  @@RD = File::expand_path(File::dirname(__FILE__) ) + File::Separator if !defined?(@@RD)
  @@config_file = "#{@@RD}config"
  @@USER_CONF_DIR =  File::join(ENV['HOME'].dup, '.cremefraiche' )
  @@USER_CONF = File::join(@@USER_CONF_DIR.dup, 'config')

  # initialize the singleton-instance and read the configuration-
  # file.	
  def initialize
    # Temporary logger. Will be replaced.
    @log = Logger.new(STDOUT)
    @log.level = Logger::INFO
    @log.info('init config')
    read_config
    check_programs_installed
    # here...
    init_logger(@conf['log_file'] ? @conf['log_file'] : STDOUT, @conf['log_level'] ? @conf['log_level'] : Logger::UNKNOWN)
  end
  def save_config(conf)
    yml = conf.to_yaml
    @log.debug('yaml will be ' << yml)
    if (! Dir.exist?(@@USER_CONF_DIR) ) 
      Dir.mkdir(@@USER_CONF_DIR)
    end
    File.open(@@USER_CONF, 'w') do |cf|
      cf << yml	
    end
  end
  def to_s
    str = @conf.to_s
    @log.debug(str)
    return str
  end

  # used to extract values from the confiuration-file
  def method_missing(method, *args)
    # @log.debug('method missing ' << method.to_s << '(' << args.join(', ') << ')')
    key = method.to_s.gsub('_', ' ')
    if(@conf && key)
      co = @conf[key]
      # @log.debug('returning conf-option ' << (co && !co.to_s.empty? ? co.to_s : 'nil'))
      return (co && !co.to_s.empty? ? co : nil)
    else
      @log.error(trl("Unable to read an option from the configuration"))
    end
  end

  # returns the path to the configuration-file
  def config_file
    @@config_file
  end

  def size
    @conf.size
  end

  def each_with_index(&b)
    @conf.each_with_index do |pair, i|
      yield(pair, i)
    end
  end

  def each_pair(&b)
    @conf.each_pair do |k, v|
      yield(k, v)
    end
  end

  # Creates a temporary directory upon the first call but always returns its path
  # Stolen from Alex Chaffee's files-gem and 
  # http://stackoverflow.com/questions/1139265/what-is-the-best-way-to-get-a-temporary-directory-in-ruby-on-rails
  def temp_dir options = {:remove => true}
    @temp_dir ||= begin
                    @log.debug('creating temp_dir')
                    require 'tmpdir'
                    require 'fileutils'
                    #
                    # called_from = File.basename caller.first.split(':').first, ".rb"
                    called_from = CremeFraiche::prog_info[:app_name].gsub(' ', '_')
                    @log.debug('called_from is ' << called_from)
                    path = File.join(Dir::tmpdir, "#{called_from}_#{Time.now.to_i}_#{rand(1000)}")
                    @log.debug('temp-dir path is ' << path)
                    Dir.mkdir(path)
                    at_exit {FileUtils.rm_rf(path) if File.exists?(path)} if options[:remove]
                    File.new path
                  end
  end

  def read_config
    # replace the globally defined configuration file by a 
    # user-provided version, if available.
    # -----> here
    @has_user_conf = user_conf
    # <------
    msg = file_check(@@config_file, :exist, :readable)	
    if(!msg) 
      @log.info('configuration will be read from ' << @@config_file)
      begin
        @conf = YAML::load_file(@@config_file)
        @keys = @conf.keys
        if(@keys.include?('log level') )	
          @log_level = case @conf['log level'].downcase
                       when 'debug'
                         Logger::DEBUG
                       when 'fatal'
                         Logger::FATAL
                       when 'error'
                         Logger::ERROR
                       when 'info'
                         Logger::INFO
                       when 'warn'
                         Logger::WARN
                       else
                         Logger::UNKNOWN
                       end
        else
          @log_level = Logger::UNKNOWN
        end
      rescue Exception => ex
        msg = ex.message
      end
      if(@conf) 
        init_logger(@conf['log file'] ? @conf['log file'] : STDOUT, @log_level)
      else
        @log.fatal('Cannot set the log-level ! @conf is ' << @conf.inspect << '. EXITING!')
        exit -1
      end
    end
    @log.debug('configuration is ' << @conf.to_s)	
    @log.error(trl("cannot read the configuration-file (%s)") %msg ) if msg

  end



  attr_reader :log_level
  attr_reader :has_user_conf
  attr_reader :keys
  attr_reader :user_configuration

  private

  # Finds external programs in the environment path.
  # A decision to use one of those programs depends on the fact, 
  # that they are installed at all. Contradicting options will 
  # result in warning-messages in the protocol output.
  def check_programs_installed
    prog_names = %w"pdftk firefox opera epiphany-browser"
    $Progs = {} if !defined? $Progs
    pathdirs = ENV['PATH'].split(File::PATH_SEPARATOR).collect do |d| 
      # BUGFIX: PATH may contain just anything!
      pdir = nil
      if(Dir.exist?(d) )
        if(File.readable?(d) )
          # create a non nil entry for pathdirs.
          begin
            pdir = Dir.new(d) 
          rescue Exception => ex
            # "Trust nobody" (Herod V)
            @log.error(trl("Cannot search for executables in %s") %d << ': ' << ex.message )
          end
        else
          @log.warn(trl('The path-variable references an unreadable directory') << ': ' << (d ? d.to_s : 'NIL') )
        end
      else
        @log.warn('test')
        # @log.warn(trl('The path-variable contains an invalid value') << ': ' << (d ? d.to_s : 'NIL') )
      end
      # --------> Now this is neat... 
      pdir
    end
    # ... or wait ... It is not!
    # Remove the nil-entries from pathdirs
    pathdirs.compact!
    # <------ Now this is neat... 

    if(pathdirs && !pathdirs.empty? )	
      @log.debug('path: ' + pathdirs.join(', ') )
      exts = ENV['PATHEXT'] ? ENV['PATHEXT'].split(';') : ['']

      prog_names.each do  |prog|
        pathdirs.each do |dir|
          exts.each do |ext|
            # BUGFIX: Beware of case-variations in the 
            # defined program-extension. Windows-specific. 
            # Thanks to Tijn Gommans for pointing this out.
            $Progs[prog] |= dir.entries.include?("#{prog}#{ext.upcase}")
            $Progs[prog] |= dir.entries.include?("#{prog}#{ext.downcase}")

          end
        end
      end
    end
    @log.debug("found the following external programs: " << $Progs.to_s)
  end


  # Checks, if a version of the configuration-file had been created in the user's
  # home-directory and switches to this file, if it exists. Otherwise the original
  # standard-version is used.
  def user_conf() 
    # does the HOME-key exist in the environment?
    home_dir = ENV['HOME']
    @log.debug('home-directory appears to be: ' << home_dir) if @log
    if(home_dir )
      # The user-provided configuration-file is $HOME/.cremefraiche/config

      if(File::exist?(@@USER_CONF) && File.readable?(@@USER_CONF) && File::file?(@@USER_CONF))
        if File::size(@@USER_CONF) > 100
          @@config_file = @@USER_CONF
          return true
        else
          @log.error('The configuration-file "' << @@USER_CONF << "\" is empty or too short!\nWill be using configuration from " << @@config_file  << '.') if @log
        end
      else
        @log.warn('Cannot use configuration-file ' << @@USER_CONF) if @log
        @log.warn('Using configuration from ' << @@config_file  << '.') if @log

      end
    end
    return false
  end


end

$CONF = Configuration.instance
